@@ -15,6 +15,10 @@ from .waldo import Waldo, Instruction, InstructionType
 
 
 # Dimensions of component types
+# TODO: Refactor so default dimensions and allowable variations on type/shape are built-in to each subclass of Component
+#       E.g. Component.DEFAULT_DIMENSIONS = (2, 3), Recycler.DEFAULT_DIMENSIONS = (5, 5)
+#       Then we should be able to avoid type string manipulations ('output' in type) and only need to specify the types
+#        that change their shape from the class' default
 COMPONENT_SHAPES = {
     # SpaceChem stores co-ordinates as col, row
     'research-input': (1, 1),
@@ -24,6 +28,7 @@ COMPONENT_SHAPES = {
     'output': (2, 3),  # All production level outputs appear to be 2x3
     'drag-recycler': (5, 5),
     'drag-storage-tank': (3, 3),
+    'drag-storage-tank-infinite': (3, 3),
     'freeform-counter': (2, 3),
     'drag-arbitrary-input': (2, 3),
     'drag-silo-input': (5, 5),
@@ -35,7 +40,9 @@ COMPONENT_SHAPES = {
     'drag-spaceship-input': (2, 2),  # TODO: Actually (2,3) but its pipe isn't in the middle which fucks our assumptions
     'drag-programmed-input': (2, 3),
     'drag-qpipe-in': (3, 1),
-    'drag-qpipe-out': (3, 1)}
+    'drag-qpipe-out': (3, 1),
+    'drag-printer-output': (2, 3),
+    'drag-printer-passthrough': (2, 3)}
 
 # Production level codes don't specify available reactor properties like research levels; encode them here
 REACTOR_TYPES = {
@@ -130,12 +137,18 @@ class Component:
             return super().__new__(Reactor)
         elif 'input' in parts:
             return super().__new__(Input)
+        elif _type == 'drag-printer-output':
+            return super().__new__(OutputPrinter)
+        elif _type == 'drag-printer-passthrough':
+            return super().__new__(PassThroughPrinter)
         elif 'output' in parts or 'production-target' in _type:
             return super().__new__(Output)
         elif _type == 'drag-recycler':
             return super().__new__(Recycler)
         elif _type == 'drag-storage-tank':
             return super().__new__(StorageTank)
+        elif _type == 'drag-storage-tank-infinite':
+            return super().__new__(InfiniteStorageTank)
         elif _type == 'freeform-counter':
             return super().__new__(PassThroughCounter)
         elif _type == 'drag-qpipe-in':
@@ -209,7 +222,7 @@ class Component:
             if not (pipe_line.startswith('PIPE:0') or pipe_line.startswith('PIPE:1')):
                 raise ValueError(f"Unexpected line in component pipes: `{pipe_line}`")
 
-        if update_pipes:
+        if update_pipes and self.out_pipes:
             # Expect the remaining lines to define the component's output pipes
             # If the pipes on an existing component are updatable, all of them must be specified during an update
             # (as testable by playing around with preset reactors in CE production levels)
@@ -241,6 +254,13 @@ class Component:
         for pipe in self.out_pipes:
             pipe.move_contents()
 
+    def export_str(self):
+        '''Represent this component in solution export string format.'''
+        # TODO: I'm still not sure what the 4th component field is used for. Custom reactor names maybe?
+
+        return f"COMPONENT:'{self.type}',{self.posn[0]},{self.posn[1]},''\n" \
+               + '\n'.join(pipe.export_str(pipe_idx=i) for i, pipe in enumerate(self.out_pipes))
+
 
 class Input(Component):
     __slots__ = 'molecules', 'input_rate'
@@ -290,11 +310,6 @@ class Input(Component):
         if (cycle - 1) % self.input_rate == 0 and self.out_pipe[0] is None:
             self.out_pipe[0] = copy.deepcopy(self.molecules[0])
 
-    def export_str(self):
-        '''Represent this input in solution export string format.'''
-        # TODO: I'm still not sure what the 4th component field is used for. Custom reactor names maybe?
-        return f"COMPONENT:'{self.type}',{self.posn[0]},{self.posn[1]},''" + '\n' + self.out_pipe.export_str()
-
 
 class RandomInput(Input):
     __slots__ = 'random_generator', 'input_counts', 'random_bucket'
@@ -337,6 +352,7 @@ class RandomInput(Input):
         if (cycle - 1) % self.input_rate == 0 and self.out_pipe[0] is None:
             self.out_pipe[0] = copy.deepcopy(self.molecules[self.get_next_molecule_idx()])
 
+
 class ProgrammedInput(Input):
     __slots__ = 'starting_molecules', 'repeating_molecules', 'repeating_idx'
 
@@ -401,10 +417,13 @@ class Output(Component):
         molecule = self.in_pipe[-1]
         if molecule is not None:
             if not molecule.isomorphic(self.output_molecule):
-                raise InvalidOutputError(f"Invalid output molecule; expected:\n{self.output_molecule}\nbut got:\n{molecule}")
+                raise InvalidOutputError(f"Invalid output molecule; expected:\n{self.output_molecule}\n\nbut got:\n{molecule}")
 
             self.in_pipe[-1] = None
-            self.current_count += 1
+
+            if self.current_count < self.target_count:
+                self.current_count += 1
+
             if self.current_count == self.target_count:
                 return True
 
@@ -475,6 +494,62 @@ class DisabledOutput(Component):
             raise InvalidOutputError("A molecule was passed to a disabled output.")
 
 
+class OutputPrinter(Component):
+    """Displays the last 3 molecules passed to it. For now this is effectively going to be a recycler..."""
+    __slots__ = ()
+
+    def __init__(self, component_dict=None, _type=None, posn=None):
+        super().__init__(component_dict, _type=_type, posn=posn, num_in_pipes=1)
+
+    @property
+    def in_pipe(self):
+        return self.in_pipes[0]
+
+    @in_pipe.setter
+    def in_pipe(self, p):
+        self.in_pipes[0] = p
+
+    def do_instant_actions(self, _):
+        """Consume and print incoming molecules."""
+        if self.in_pipe is not None:
+            # TODO: Print received molecules when in --debug somehow
+            self.in_pipe[-1] = None
+
+
+class PassThroughPrinter(OutputPrinter):
+    """Displays the last 3 molecules passed to it and passes them on."""
+    __slots__ = 'stored_molecule',
+
+    def __init__(self, component_dict=None, _type=None, posn=None):
+        super(OutputPrinter, self).__init__(component_dict, _type=_type, posn=posn, num_in_pipes=1, num_out_pipes=1)
+
+        self.stored_molecule = None
+
+    @property
+    def out_pipe(self):
+        return self.out_pipes[0]
+
+    @out_pipe.setter
+    def out_pipe(self, p):
+        self.out_pipes[0] = p
+
+    def do_instant_actions(self, cycle):
+        '''Check for and process any incoming molecule, and return True if this output just completed (in which case
+        the caller should check if the other outputs are also done). This avoids checking all output counts every cycle.
+        '''
+        if self.in_pipe is None:
+            return
+
+        # If there is a molecule stored (possibly stored just now), put it in the output pipe if possible
+        if self.stored_molecule is not None and self.out_pipe[0] is None:
+            self.stored_molecule, self.out_pipe[0] = None, self.stored_molecule
+
+        # If the stored slot is empty, store the next molecule and 'print' it while we do so
+        if self.in_pipe[-1] is not None and self.stored_molecule is None:
+            self.stored_molecule = self.in_pipe[-1]
+            super().do_instant_actions(cycle)  # This will consume and print the input molecule
+
+
 class Recycler(Component):
     __slots__ = ()
 
@@ -546,6 +621,10 @@ class StorageTank(Component):
         return cls(component_type, component_posn, out_pipe=Pipe.from_export_str(pipe_str))
 
 
+class InfiniteStorageTank(StorageTank):
+    MAX_CAPACITY = math.inf
+
+
 class TeleporterInput(Component):
     __slots__ = 'destination',
 
@@ -615,12 +694,12 @@ class Reactor(Component):
     walls = {UP: -0.5 + ATOM_RADIUS, DOWN: 7.5 - ATOM_RADIUS,
              LEFT: -0.5 + ATOM_RADIUS, RIGHT: 9.5 - ATOM_RADIUS}
     # Names of features as stored in attributes
-    FEATURE_NAMES = ('bonders', 'sensors', 'fusers', 'splitters', 'swappers', 'bonder_pluses', 'bonder_minuses')
+    FEATURE_NAMES = ('bonders', 'sensors', 'fusers', 'splitters', 'swappers')
 
     __slots__ = ('in_pipes', 'out_pipes',
                  'waldos', 'molecules',
                  'large_output', 'bonders', 'sensors', 'fusers', 'splitters', 'swappers',
-                 'bonder_pluses', 'bonder_minuses', 'bond_plus_pairs', 'bond_minus_pairs',
+                 'bond_plus_pairs', 'bond_minus_pairs',
                  'quantum_walls_x', 'quantum_walls_y', 'disallowed_instrs',
                  'debug')
 
@@ -646,13 +725,20 @@ class Reactor(Component):
 
         # Place all features
         cur_col = 0  # For simplicity we will put each feature type in its own column(s)
-        for attr_name, feature_name, feature_width, default_count in (('bonders', 'bonder', 1, None),
-                                                                      ('sensors', 'sensor', 1, 1),
+
+        # Place bonders. Different bonder types go in the same struct so they can share a priority index
+        self.bonders = []
+        for feature_name, abbrev in (('bonder', '+-'), ('bonder-plus', '+'), ('bonder-minus', '-')):
+            if f'{feature_name}-count' in component_dict:
+                self.bonders.extend([(Position(cur_col, i), abbrev)
+                                     for i in range(component_dict[f'{feature_name}-count'])])
+            cur_col += 1
+
+        # Place remaining features
+        for attr_name, feature_name, feature_width, default_count in (('sensors', 'sensor', 1, 1),
                                                                       ('fusers', 'fuser', 2, 1),
                                                                       ('splitters', 'splitter', 2, 1),
-                                                                      ('swappers', 'teleporter', 1, 2),
-                                                                      ('bonder_pluses', 'bonder-plus', 1, None),
-                                                                      ('bonder_minuses', 'bonder-minus', 1, None),):
+                                                                      ('swappers', 'teleporter', 1, 2)):
             if f'{feature_name}-count' in component_dict:
                 setattr(self, attr_name, [Position(cur_col, i) for i in range(component_dict[f'{feature_name}-count'])])
             elif f'has-{feature_name}' in component_dict and component_dict[f'has-{feature_name}']:
@@ -706,36 +792,25 @@ class Reactor(Component):
         self.molecules = {}
 
     def bond_pairs(self):
-        '''For each of + and - bond commands, return a list of (bonder_A_posn, bonder_B_posn, dirn) triplets,
+        '''For each of + and - bond commands, return a tuple of (bonder_A_posn, bonder_B_posn, dirn) triplets,
         sorted in priority order.
         '''
-        # TODO: SC respects the priority order that bond+ vs bond- vs regular bonders are defined in in the solution
-        #       string. We need to do the same here; currently regular bonders are always ending up higher priority
+        pair_lists = []
+        for bond_type in ('+', '-'):
+            # Store the relevant types of bonders in a dict paired up with their indices for fast lookup/sorting (below)
+            bonders = {posn: i for i, (posn, bond_types) in enumerate(self.bonders) if bond_type in bond_types}
+            pair_lists.append(tuple((posn, neighbor_posn, direction)
+                                     for posn in bonders
+                                     for neighbor_posn, direction in
+                                     sorted([(posn + direction, direction)
+                                             for direction in (RIGHT, DOWN)
+                                             if posn + direction in bonders],
+                                            key=lambda x: bonders[x[0]])))
 
-        # Store the relevant types of bonders in a dict paired up with their indices for fast lookup/sorting (below)
-        bond_plus_bonders = {posn: i for i, posn in enumerate(self.bonders + self.bonder_pluses)}
-        bond_plus_pairs = tuple((posn, neighbor_posn, direction)
-                                 for posn in bond_plus_bonders
-                                 for neighbor_posn, direction in
-                                 sorted([(posn + direction, direction)
-                                         for direction in (RIGHT, DOWN)
-                                         if posn + direction in bond_plus_bonders],
-                                        key=lambda x: bond_plus_bonders[x[0]]))
-
-        bond_minus_bonders = {posn: i for i, posn in enumerate(self.bonders + self.bonder_minuses)}
-        bond_minus_pairs = tuple((posn, neighbor_posn, direction)
-                                  for posn in bond_minus_bonders
-                                  for neighbor_posn, direction in
-                                  sorted([(posn + direction, direction)
-                                          for direction in (RIGHT, DOWN)
-                                          if posn + direction in bond_minus_bonders],
-                                         key=lambda x: bond_minus_bonders[x[0]]))
-
-        return bond_plus_pairs, bond_minus_pairs
+        return pair_lists
 
     def update_from_export_str(self, export_str, update_pipes=True):
-        features = {'bonders':[], 'sensors': [], 'fusers': [], 'splitters': [], 'swappers': [],
-                    'bonder_pluses': [], 'bonder_minuses': []}
+        features = {'bonders': [], 'sensors': [], 'fusers': [], 'splitters': [], 'swappers': []}
 
         # One map for each waldo, of positions to pairs of arrows (directions) and/or non-arrow instructions
         # TODO: usage might be cleaner if separate arrow_maps and instr_maps... but probably more space
@@ -792,7 +867,7 @@ class Reactor(Component):
                     feature_posns.add(position2)
 
                 if member_name == 'feature-bonder':
-                    features['bonders'].append(position)
+                    features['bonders'].append((position, '+-'))
                 elif member_name == 'feature-sensor':
                     features['sensors'].append(position)
                 elif member_name == 'feature-fuser':
@@ -802,9 +877,9 @@ class Reactor(Component):
                 elif member_name == 'feature-tunnel':
                     features['swappers'].append(position)
                 elif member_name == 'feature-bonder-plus':
-                    features['bonder_pluses'].append(position)
+                    features['bonders'].append((position, '+'))
                 elif member_name == 'feature-bonder-minus':
-                    features['bonder_minuses'].append(position)
+                    features['bonders'].append((position, '-'))
                 else:
                     raise Exception(f"Unrecognized member type {member_name}")
 
@@ -881,6 +956,16 @@ class Reactor(Component):
 
         self.waldos = [Waldo(idx=i, instr_map=waldo_instr_maps[i]) for i in range(self.NUM_WALDOS)]
 
+        # Since bonders of different types get stored together to share a priority idx, check their individual counts
+        # match the existing counts
+        for bonder_type, feature_name in (('+-', 'bonders'),
+                                          ('+', 'bonder-pluses'),
+                                          ('-', 'bonder-minuses')):
+            actual_count = sum(1 for _, bt in features['bonders'] if bt == bonder_type)
+            expected_count = sum(1 for _, bt in self.bonders if bt == bonder_type)
+            assert actual_count == expected_count, \
+                f"Expected {expected_count} {feature_name} for {self.type} reactor but got {actual_count}"
+
         # Sanity-check and set features
         for feature_name, posns in features.items():
             assert len(posns) == len(getattr(self, feature_name)), \
@@ -894,20 +979,25 @@ class Reactor(Component):
         export_str = f"COMPONENT:'{self.type}',{self.posn.col},{self.posn.row},''"
 
         # TODO: Make reactors more agnostic of feature types
-        for posn in self.bonders:
-            export_str += f"\nMEMBER:'feature-bonder',-1,0,1,{posn.col},{posn.row},0,0"
+        for (posn, bond_types) in self.bonders:
+            if bond_types == '+-':
+                feature_name = 'bonder'
+            elif bond_types == '+':
+                feature_name = 'bonder-plus'
+            elif bond_types == '-':
+                feature_name = 'bonder-minus'
+            else:
+                raise Exception("Invalid bonder type in internal data")
+
+            export_str += f"\nMEMBER:'feature-{feature_name}',-1,0,1,{posn.col},{posn.row},0,0"
         for posn in self.sensors:
             export_str += f"\nMEMBER:'feature-sensor',-1,0,1,{posn.col},{posn.row},0,0"
         for posn in self.fusers:
             export_str += f"\nMEMBER:'feature-fuser',-1,0,1,{posn.col},{posn.row},0,0"
         for posn in self.splitters:
             export_str += f"\nMEMBER:'feature-splitter',-1,0,1,{posn.col},{posn.row},0,0"
-        for posn in self.splitters:
+        for posn in self.swappers:
             export_str += f"\nMEMBER:'feature-tunnel',-1,0,1,{posn.col},{posn.row},0,0"
-        for posn in self.bonder_pluses:
-            export_str += f"\nMEMBER:'feature-bonder-plus',-1,0,1,{posn.col},{posn.row},0,0"
-        for posn in self.bonder_minuses:
-            export_str += f"\nMEMBER:'feature-bonder-minus',-1,0,1,{posn.col},{posn.row},0,0"
 
         export_str += '\n' + '\n'.join(waldo.export_str() for waldo in self.waldos)
         export_str += '\n' + '\n'.join(pipe.export_str(pipe_idx=i) for i, pipe in enumerate(self.out_pipes))
@@ -991,7 +1081,10 @@ class Reactor(Component):
                     cell[1] = atom.element.symbol[1]
 
         # Use grey background for feature cells (bonders, fusers, etc.)
-        feature_colors = {k: 'light_slate_grey' for k in Reactor.FEATURE_NAMES}
+        # Though they're stored together for priority reasons in Reactor.bonders, color + and - bonders separately
+        # from regular bonders so as to distinguish them during + vs - commands
+        feature_colors = {k: 'light_slate_grey'
+                          for k in list(Reactor.FEATURE_NAMES) + ['bonder_pluses', 'bonder_minuses']}
         input_colors = {}
         output_colors = {}
 
@@ -1026,7 +1119,15 @@ class Reactor(Component):
 
         # Color background of feature cells
         for feature_name, feature_color in feature_colors.items():
-            cell_posns = set(getattr(self, feature_name))
+            # Extract individual bonder types from their shared struct
+            if feature_name == 'bonders':
+                cell_posns = set(p for p, bond_types in self.bonders if bond_types == '+-')
+            elif feature_name == 'bonder_pluses':
+                cell_posns = set(p for p, bond_types in self.bonders if bond_types == '+')
+            elif feature_name == 'bonder_minuses':
+                cell_posns = set(p for p, bond_types in self.bonders if bond_types == '-')
+            else:
+                cell_posns = set(getattr(self, feature_name))
 
             for c, r in cell_posns:
                 cells[r][c][0] = f'[on {feature_color}]{cells[r][c][0]}[/]'
@@ -1391,10 +1492,10 @@ class Reactor(Component):
             if direction not in atom_A.bonds or atom_A.bonds[direction] != 3:
                 atom_B = molecule_B[neighbor_posn]
 
-                # Do nothing if either atom is at its bond limit (spacechem does not mark
-                # any molecules as modified in this case unless the bond was size 3)
-                if (sum(atom_A.bonds.values()) == atom_A.element.max_bonds
-                        or sum(atom_B.bonds.values()) == atom_B.element.max_bonds):
+                # Do nothing if either atom is at (or above) its bond limit (spacechem does not mark any molecules as
+                # modified in this case unless the bond was size 3)
+                if (sum(atom_A.bonds.values()) >= atom_A.element.max_bonds
+                        or sum(atom_B.bonds.values()) >= atom_B.element.max_bonds):
                     continue
 
                 direction_B = direction.opposite()
