@@ -27,7 +27,7 @@ IS_WINDOWS = platform.system() == 'Windows'
 class DebugOptions:
     """Debug options for running a solution.
 
-    reactor: The reactor to debug. If None, overworld is shown for production levels, or only reactor for researches
+    reactor: The reactor to debug. If None, overworld is shown for production levels, or only reactor for researches.
     cycle: The cycle to start debugging at. Default 0.
     speed: The speed to run the solution at, in cycles / s. Default 10.
     show_instructions: Whether to show waldo instructions. Default False as this can actually reduce readability.
@@ -53,7 +53,7 @@ class Score(namedtuple("Score", ('cycles', 'reactors', 'symbols'))):
         '''
         parts = s.split('-')
         return (len(parts) == 3
-                and (parts[0].isdigit() or parts[0] == 'Incomplete') # Thinks 045-1-14 is a score but whatever
+                and (parts[0].isdigit() or parts[0] == 'Incomplete')  # Thinks 045-1-14 is a score but whatever
                 and all(part.isdigit() for part in parts[1:]))
 
     @classmethod
@@ -89,6 +89,10 @@ class Solution:
     def outputs(self):
         return (component for component in self.components if isinstance(component, Output))
 
+    @property
+    def symbols(self):
+        return sum(sum(len(waldo) for waldo in reactor.waldos) for reactor in self.reactors)
+
     @property
     def description(self):
         return self.describe(self.level.name, self.author, self.expected_score, self.name)
@@ -164,7 +168,9 @@ class Solution:
         self.cycle = 0
 
         # Set up the level terrain so we can look up input/output positions and any blocked terrain
-        if level['type'].startswith('production'):
+        if level['type'].startswith('research'):
+            terrain_id = 'research'
+        elif level['type'].startswith('production'):
             # Main game levels have unique terrain which we have to hardcode D:
             # We can at least avoid name collisions if we deliberately don't add the terrain field to their JSONs
             # TODO: This is a bit dangerous; the game auto-adds a terrain field if it ever gets its hands on the json
@@ -174,8 +180,10 @@ class Solution:
             # SC caps oversized terrain IDs
             if isinstance(terrain_id, int) and terrain_id > MAX_TERRAIN_INT:
                 terrain_id = MAX_TERRAIN_INT
+        elif level['type'] == 'sandbox':
+            terrain_id = 7
         else:
-            terrain_id = 'research'  # BIG HACKS
+            raise ValueError(f"Unrecognized level type {repr(level['type'])}")
 
         # Add level-defined entities - including any fixed-position reactors/pipes
         # Store components by posn for now, for convenience. We'll re-order and store them in self.components later
@@ -194,7 +202,9 @@ class Solution:
             input_zone_types = ('random-input-components', 'programmed-input-components')
             output_zone_type = 'output-components'
         else:
-            raise ValueError(f"Unknown level type {self.level['type']}")
+            # Sandbox
+            input_zone_types = ('random-input-zones', 'fixed-input-zones')
+            output_zone_type = None
 
         for input_zone_type in input_zone_types:
             if input_zone_type not in self.level:
@@ -233,23 +243,37 @@ class Solution:
                     new_component = Input(input_dict=input_dict, is_research=self.level['type'].startswith('research'))
                     posn_to_component[new_component.posn] = new_component
 
+        # Sandboxes use an alternate format for specifying the programmed input
+        # Convert to the standard CE-format programmed input dict
+        if self.level['type'] == 'sandbox' and 'programmed-input-molecules' in self.level:
+            input_dict = {'starting-molecules': [self.level['programmed-input-molecules'][str(i)]
+                                                 for i in self.level['programmed-input-start']],
+                          'repeating-molecules': [self.level['programmed-input-molecules'][str(i)]
+                                                  for i in self.level['programmed-input-repeat']]}
+            component_type, component_posn = terrains[terrain_id]['programmed-input']
+
+            posn_to_component[component_posn] = Input(input_dict=input_dict,
+                                                      _type=component_type, posn=component_posn,
+                                                      is_research=False)
+
         # Outputs
-        if isinstance(self.level[output_zone_type], dict):
-            for i, output_dict in self.level[output_zone_type].items():
-                i = int(i)
+        if output_zone_type is not None:
+            if isinstance(self.level[output_zone_type], dict):
+                for i, output_dict in self.level[output_zone_type].items():
+                    i = int(i)
 
-                # I'd handle this in the Output constructor but I refuse to enable Zach's madness
-                if self.level['type'] == 'production':
-                    output_dict = copy.deepcopy(output_dict)  # To avoid mutating the level
-                    output_dict['count'] *= 4  # Zach pls
+                    # I'd handle this in the Output constructor but I refuse to enable Zach's madness
+                    if self.level['type'] == 'production':
+                        output_dict = copy.deepcopy(output_dict)  # To avoid mutating the level
+                        output_dict['count'] *= 4  # Zach pls
 
-                component_type, component_posn = terrains[terrain_id][output_zone_type][i]
-                posn_to_component[component_posn] = Output(output_dict=output_dict,
-                                                           _type=component_type, posn=component_posn)
-        else:
-            for output_dict in self.level[output_zone_type]:
-                new_component = Output(output_dict=output_dict)
-                posn_to_component[new_component.posn] = new_component
+                    component_type, component_posn = terrains[terrain_id][output_zone_type][i]
+                    posn_to_component[component_posn] = Output(output_dict=output_dict,
+                                                               _type=component_type, posn=component_posn)
+            else:
+                for output_dict in self.level[output_zone_type]:
+                    new_component = Output(output_dict=output_dict)
+                    posn_to_component[new_component.posn] = new_component
 
         # Add disabled output components for the unused outputs of research levels (crash if given a molecule)
         if self.level['type'].startswith('research') and not ('has-large-output' in self.level
@@ -322,7 +346,12 @@ class Solution:
                     assert ((reactor_type_flag in self.level and self.level[reactor_type_flag])
                             # Misleadingly, allowed-reactor-types includes storage tanks
                             or ('allowed-reactor-types' in self.level
-                                and component_type in self.level['allowed-reactor-types'])), \
+                                and component_type in self.level['allowed-reactor-types'])
+                            # Sandbox levels always allow sandbox reactor, infinite storage tank, and printers
+                            or (self.level['type'] == 'sandbox' and component_type in ('drag-sandbox-reactor',
+                                                                                       'drag-storage-tank-infinite',
+                                                                                       'drag-printer-passthrough',
+                                                                                       'drag-printer-output'))), \
                         f"New component type {component_type} (at {component_posn}) is not legal in this level"
 
                     component_dict = {}
@@ -363,7 +392,7 @@ class Solution:
                     component.update_from_export_str(component_str, update_pipes=update_pipes)
                 except Exception as e:
                     if not self.level['type'].startswith('research'):
-                        raise type(e)(f"{component.type} at {component.posn}: {e}")
+                        raise type(e)(f"{component.type} at {component.posn}: {e}") from e
 
                     raise e
 
@@ -499,14 +528,11 @@ class Solution:
                 export_str += f",{self.name}"
 
         # Components
-        # Exclude inputs whose pipes are length 1 (unmodified), outputs, and the recycler.
-        # I'm probably forgetting another case.
+        # Exclude inputs whose pipes are length 1 (unmodified), and out-pipeless components like outputs and recycler
         # TODO: This doesn't cover inputs with preset pipes > 1 long - which also shouldn't be included
-        #       Really it's probably just every unmodified preset component
+        #       Really it's probably just every unmodified preset component.
         for component in self.components:
-            if not (isinstance(component, (Output, Recycler))
-                    or (isinstance(component, Input)
-                        and len(component.out_pipe) == 1)):
+            if component.out_pipes and not (isinstance(component, Input) and len(component.out_pipe) == 1):
                 export_str += '\n' + component.export_str()
 
         return export_str
@@ -603,8 +629,14 @@ class Solution:
                             raise type(e)(f"Reactor {i}: {e}") from e
                 raise e
 
-    def run(self, max_cycles=None, debug=False):
-        '''Run this solution, returning a score tuple or else raising an exception if the level was not solved.'''
+    def run(self, max_cycles=None, debug: Optional[DebugOptions] = False):
+        '''Run this solution, returning a Score or else raising an exception if the level was not solved.
+
+        Args:
+            max_cycles: Maximum cycle count to run to. Default double the expected cycle count in the solution metadata,
+                        or 1,000,000 cycles if not provided (use math.inf if you don't fear infinite loop solutions).
+            debug: Print an updating view of the solution while running. See DebugOptions.
+        '''
         # TODO: Running the solution should not meaningfully modify it. Namely, need to reset reactor.molecules
         #       and waldo.position/waldo.direction before each run, or otherwise prevent these from persisting.
         #       Otherwise a solution will only be safely runnable once.
@@ -627,6 +659,7 @@ class Solution:
         reactors = list(self.reactors)
         outputs = list(self.outputs)
 
+        # If we are continuing running from a pause, start with a movement phase
         if any(waldo.instr_map[waldo.position][1].type == InstructionType.PAUSE
                for reactor in reactors
                for waldo in reactor.waldos
@@ -634,10 +667,6 @@ class Solution:
             self.cycle_movement()
 
         # Run the level
-        symbols = sum(sum(len(waldo) for waldo in component.waldos)
-                      for component in self.components
-                      if hasattr(component, 'waldos'))  # hacky but saves on using a counter or reactor list in the ctor
-
         try:
             # In debug mode the cursor can annoyingly flicker into the middle of the printed output; hide it
             if debug:
@@ -657,7 +686,7 @@ class Solution:
                             # TODO: Update solution expected score? That would match the game's behavior, but makes the validator
                             #       potentially misleading. Maybe run() and validate() should be the same thing.
                             # -1 looks weird but seems provably right based on output vs pause comparison
-                            return Score(self.cycle - 1, len(reactors), symbols)
+                            return Score(self.cycle - 1, len(reactors), self.symbols)
 
                 if debug and self.cycle >= debug.cycle:
                     self.debug_print(duration=0.5 / debug.speed, reactor_idx=debug.reactor,
@@ -703,9 +732,7 @@ class Solution:
             print(f"Warning: Validating solution against level {repr(self.level.name)} that was originally"
                   + f" constructed for level {repr(self.level_name)}.")
 
-        if max_cycles is None:
-            max_cycles = 2 * self.expected_score.cycles
-        elif self.expected_score.cycles > max_cycles:
+        if max_cycles is not None and self.expected_score.cycles > max_cycles:
             raise ValueError(f"{self.description}: Cannot validate; expected cycles > max cycles ({max_cycles})")
 
         score = self.run(max_cycles=max_cycles, debug=debug)
